/*
 * Copyright (c) 2011, Olivier MATZ <zer0@droids-corp.org>
 * All rights reserved.
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the University of California, Berkeley nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE REGENTS AND CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#define _BSD_SOURCE /* for be64toh, endian.h */

#include <stdio.h>
#include <string.h>
#include <stdint.h>
#include <inttypes.h>
#include <stdlib.h>
#include <stdarg.h>
#include <errno.h>
#include <termios.h>
#include <ctype.h>
#include <unistd.h>
#include <endian.h>
#include <time.h>
#include <netinet/in.h>
#include <sys/queue.h>
#include <linux/joystick.h>

#include <getopt.h>

#include <cmdline_rdline.h>
#include <cmdline_parse.h>
#include <cmdline_socket.h>
#include <cmdline.h>

#include <event.h>

#include "xbee_neighbor.h"
#include "xbee_atcmd.h"
#include "xbee_stats.h"
#include "xbee_buf.h"
#include "xbee_proto.h"
#include "xbee.h"
#include "joystick.h"
#include "rc_proto.h"
#include "main.h"

#define TIMEOUT_US 1000000

/* XXX qd on deconnecte, prend 100% du cpu */
/* XXX rmt at cmd */
/* XXX neighbor discovery */
/* XXX dump et restauration de la config */

/* global xbee device */
static int xbeeapp_dump_conf_continue(struct xbee_ctx *ctx,
				      struct xbee_atresp_hdr *frame,
				      unsigned len);

struct xbee_dev *xbee_dev;

/* events */
static struct event stdin_read_event, xbee_read_event;
static struct event joystick_read_event;

static struct cmdline *xbee_cl;

/* parameters */
static char *xbee_devname = NULL;
static int xbee_baud = 9600;
int xbee_raw = 0;
int xbee_hexdump = 0;
int xbee_debug = 0;
FILE *xbee_logfile;
static char *joystick_devname = NULL;
struct joystick_info joyinfo;

void xbeeapp_log(int always_on_stdout, const char *fmt, ...)
{
	va_list ap, ap2;

	va_start(ap, fmt);
	va_copy(ap2, ap);
	if (xbee_logfile != NULL)
		vfprintf(xbee_logfile, fmt, ap);
	if (always_on_stdout || xbee_debug)
		vprintf(fmt, ap2);
	va_end(ap2);
	va_end(ap);
}

static void hexdump(int always_on_stdout, const char *title, const void *buf,
		    unsigned int len)
{
	unsigned int i, out, ofs;
	const unsigned char *data = buf;
#define LINE_LEN 80
	char line[LINE_LEN];	/* space needed 8+16*3+3+16 == 75 */
	struct timeval tv;

	gettimeofday(&tv, NULL);

	xbeeapp_log(always_on_stdout, "%"PRIi64".%.6d %s at [%p], len=%d\n",
		    (uint64_t)tv.tv_sec, (int)tv.tv_usec, title, data, len);
	ofs = 0;
	while (ofs < len) {
		/* format 1 line in the buffer, then use printk to print them */
		out = snprintf(line, LINE_LEN, "%08X", ofs);
		for (i=0; ofs+i < len && i<16; i++)
			out += snprintf(line+out, LINE_LEN - out, " %02X",
					data[ofs+i]&0xff);
		for (;i<=16;i++)
			out += snprintf(line+out, LINE_LEN - out, "   ");
		for (i=0; ofs < len && i<16; i++, ofs++) {
			unsigned char c = data[ofs];
			if (!isascii(c) || !isprint(c))
				c = '.';
			out += snprintf(line+out, LINE_LEN - out, "%c", c);
		}
		xbeeapp_log(always_on_stdout, "%s\n", line);
	}
}

static int parse_xmit_status(struct xbee_ctx *ctx,
			     struct xbee_xmit_status_hdr *frame, unsigned len)
{
	if (ctx == NULL) {
		xbeeapp_log(1, "no context\n");
		return -1;
	}

	/* see if it matches a xmit query */
	if (ctx->type != SEND_MSG) {
		xbeeapp_log(ctx->foreground, "invalid response\n");
		return -1;
	}

	/* XXX use defines for these values */
	if (frame->delivery_status == 0x00)
		xbeeapp_log(ctx->foreground, "Success\n");
	else if (frame->delivery_status == 0x01)
		xbeeapp_log(ctx->foreground, "MAC ACK Failure\n");
	else if (frame->delivery_status == 0x15)
		xbeeapp_log(ctx->foreground, "Invalid destination endpoint\n");
	else if (frame->delivery_status == 0x21)
		xbeeapp_log(ctx->foreground, "Network ACK Failure\n");
	else if (frame->delivery_status == 0x25)
		xbeeapp_log(ctx->foreground, "Route Not Found\n");

	return 0;
}

static int atcmd_frame_status(struct xbee_atresp_hdr *frame, unsigned len)
{
	char atcmd_str[3];

	/* get AT command from frame */
	memcpy(atcmd_str, &frame->cmd, 2);
	atcmd_str[2] = '\0';

	if (frame->status == 1)
		xbeeapp_log(1, "<%s>: Status is error\n", atcmd_str);
	else if (frame->status == 2)
		xbeeapp_log(1, "<%s>: Invalid command\n", atcmd_str);
	else if (frame->status == 3)
		xbeeapp_log(1, "<%s>: Invalid parameter\n", atcmd_str);
	else if (frame->status != 0)
		xbeeapp_log(1, "<%s>: Unknown status error %d\n", atcmd_str,
		       frame->status);
	return frame->status;
}

/* assume frame->status is 0 */
static int atcmd_frame_to_string(struct xbee_atresp_hdr *frame, unsigned len,
				 char *dstbuf, unsigned dstlen)
{
	union {
		uint8_t u8;
		uint16_t u16;
		uint32_t u32;
		int16_t s16;
	} __attribute__((packed)) *result;
	char atcmd_str[3];
	struct xbee_atcmd *cmd;

	/* get AT command from frame */
	memcpy(atcmd_str, &frame->cmd, 2);
	atcmd_str[2] = '\0';

	/* see if it exists */
	cmd = xbee_atcmd_lookup_name(atcmd_str);
	if (cmd == NULL) {
		xbeeapp_log(1, "unknown response\n");
		return -1;
	}

	/* dump frame */
	result = (void *)frame->data;
	len -= offsetof(struct xbee_atresp_hdr, data);
	if (cmd->flags & XBEE_ATCMD_F_PARAM_NONE && len == 0)
		snprintf(dstbuf, dstlen, "ok");
	if (cmd->flags & XBEE_ATCMD_F_PARAM_U8 && len == sizeof(uint8_t))
		snprintf(dstbuf, dstlen, "0x%x", result->u8);
	else if (cmd->flags & XBEE_ATCMD_F_PARAM_U16 && len == sizeof(uint16_t))
		snprintf(dstbuf, dstlen, "0x%x", ntohs(result->u16));
	else if (cmd->flags & XBEE_ATCMD_F_PARAM_U32 && len == sizeof(uint32_t))
		snprintf(dstbuf, dstlen, "0x%x", ntohl(result->u32));
	else if (cmd->flags & XBEE_ATCMD_F_PARAM_S16 && len == sizeof(int16_t))
		snprintf(dstbuf, dstlen, "%d\n", ntohs(result->s16));
	else
		return -1;

	return 0;
}

static int dump_atcmd(struct xbee_ctx *ctx, struct xbee_atresp_hdr *frame,
		      unsigned len)
{
	char buf[32];

	/* see if it matches query */
	if (memcmp(&frame->cmd, ctx->atcmd_query->name, 2)) {
		printf("invalid response\n");
		return -1;
	}

	/* dump frame */
	if (atcmd_frame_status(frame, len) == 0) {

		/* callback */
		if (ctx->func != NULL)
			ctx->func(frame, len, ctx->arg);

		if (len == sizeof(struct xbee_atresp_hdr))
			xbeeapp_log(ctx->foreground, "<%s>: ok\n",
				    ctx->atcmd_query->name);
		else if (atcmd_frame_to_string(frame, len, buf,
					      sizeof(buf)) == 0)
			xbeeapp_log(ctx->foreground, "<%s> is %s\n",
				    ctx->atcmd_query->name, buf);
		else
			hexdump(ctx->foreground, "atcmd answer", frame, len);
	}
	return 0;
}

int xbee_recv_data(struct xbee_recv_hdr *recvframe, unsigned len)
{
	int datalen = len - sizeof(*recvframe);
	struct rc_proto_hdr *rch = (struct rc_proto_hdr *) &recvframe->data;

	if (datalen < sizeof(struct rc_proto_hdr))
		return -1;

	switch (rch->type) {
		case RC_PROTO_TYPE_CHANNEL:
			if (datalen != sizeof(struct rc_proto_channel))
				return -1;
			break;
		case RC_PROTO_TYPE_RANGE: {
			struct rc_proto_range *rcr =
				(struct rc_proto_range *) recvframe->data;

			if (datalen != sizeof(struct rc_proto_range))
				return -1;

			if (rcr->power_level >= MAX_POWER_LEVEL)
				return -1;

			rc_proto_rx_range(rcr->power_level);

			break;
		}
		default:
			return -1;
	}

	return 0;
}


void xbee_rx(struct xbee_dev *dev, int channel, int type,
	     void *frame, unsigned len, void *opaque)
{
	struct xbee_ctx *ctx = opaque;
	int do_hexdump = xbee_hexdump;

	xbeeapp_log(0, "type=0x%x, channel=%d, ctx=%p\n", type, channel, ctx);
	hexdump(0, "rx", frame, len);

	/* if ctx is !NULL, it is an answer to a query */
	if (ctx != NULL) {
		/* XXX only delete timeout if answer matched */
		xbee_unload_timeout(ctx);
		if (xbee_debug && ctx->atcmd_query != NULL)
			printf("Received answer to query <%s>\n",
			       ctx->atcmd_query->name);
		xbee_unregister_channel(dev, channel);
	}

	/* some additional checks before sending */
	switch (type) {
		case XBEE_TYPE_MODEM_STATUS: {
			printf("Received Modem Status frame\n");
			break;
		}

		case XBEE_TYPE_RMT_ATRESP: {
			uint64_t addr;
			memcpy(&addr, frame, sizeof(addr));
			addr = be64toh(addr);
			printf("from remote address %"PRIx64"\n", addr);

			/* this answer contains an atcmd answer at offset 10 */
			if (ctx != NULL && ctx->type == ATCMD_RMT) {
				if (dump_atcmd(ctx, frame + 10, len - 10) < 0)
					do_hexdump = 1;
			}
			else {
				printf("invalid response\n");
				do_hexdump = 1;
			}
			break;
		}
		case XBEE_TYPE_ATRESP: {
			/* we are currently dumping config, continue */
			if (ctx != NULL && ctx->type == DUMP_CONF) {
				xbeeapp_dump_conf_continue(ctx, frame, len);
			}
			else if (ctx != NULL && ctx->type == ATCMD) {
				if (dump_atcmd(ctx, frame, len) < 0)
					do_hexdump = 1;
			}
			else {
				printf("invalid response\n");
				do_hexdump = 1;
			}
			break;
		}

		case XBEE_TYPE_XMIT_STATUS: {
			if (parse_xmit_status(ctx, frame, len) < 0)
				do_hexdump = 1;
			break;
		}

		case XBEE_TYPE_RECV: {
			if (xbee_recv_data(frame, len) < 0)
				do_hexdump = 1;
			break;
		}

		case XBEE_TYPE_ATCMD:
		case XBEE_TYPE_ATCMD_Q:
		case XBEE_TYPE_XMIT:
		case XBEE_TYPE_EXPL_XMIT:
		case XBEE_TYPE_RMT_ATCMD:
		case XBEE_TYPE_EXPL_RECV:
		case XBEE_TYPE_NODE_ID:
		default:
			xbeeapp_log(0, "Invalid frame\n");
			do_hexdump = 1;
			break;
	}

	if (do_hexdump) {
		xbeeapp_log(1, "rx frame type=0x%x, channel=%d, ctx=%p\n",
			    type, channel, ctx);
		hexdump(1, "undecoded rx frame", frame, len);
	}

	/* restart command line if it was a blocking query */
	if (ctx != NULL) {
		if (ctx->foreground && !ctx->have_more_command) {
			xbee_stdin_enable();
			rdline_newline(&xbee_cl->rdl, xbee_cl->prompt);
		}
		if (!ctx->have_more_command)
			free(ctx);
	}
}

static int xbeeapp_send(struct xbee_ctx *ctx, int type, void *buf, unsigned len,
			int foreground)
{
	int ret;
	int channel;

	if (len > XBEE_MAX_FRAME_LEN) {
		printf("frame too large\n");
		return -1;
	}

	/* register a channel */
	channel = xbee_register_channel(xbee_dev, XBEE_CHANNEL_ANY,
					xbee_rx, ctx);
	if (channel < 0) {
		printf("cannot send: no free channel\n");
		return -1;
	}

	xbeeapp_log(0, "send frame ctx=%p channel=%d type=0x%x len=%d\n",
	       ctx, channel, type, len);
	hexdump(0, "xmit frame", buf, len);

	/* transmit the frame on this channel */
	ret = xbee_proto_xmit(xbee_dev, channel, type, buf,
			      len);
	if (ret < 0) {
		printf("cannot send: %s\n", strerror(errno));
		xbee_unregister_channel(xbee_dev, channel);
		return -1;
	}

	ctx->channel = channel;
	xbee_load_timeout(ctx);   /* load a timeout event */

	/* suspend command line until we have answer or timeout */
	if (foreground) {
		ctx->foreground = 1;
		rdline_stop(&xbee_cl->rdl); /* don't display prompt when return */
		xbee_stdin_disable();  /* unload file descriptor polling */
	}

	return 0;
}

/* send an AT command with parameters filled by caller. Disable
 * command line until we get the answer or until a timeout occurs */
int xbeeapp_send_atcmd(const char *atcmd_str,
		       void *param, unsigned param_len, int foreground,
		       int (*func)(void *frame, unsigned len, void *arg),
		       void *arg)
{
	struct xbee_ctx *ctx;
	struct xbee_atcmd *cmd;
	struct {
		struct xbee_atcmd_hdr atcmd;
		char buf[XBEE_MAX_FRAME_LEN];
	} __attribute__((packed)) frame;

	cmd = xbee_atcmd_lookup_name(atcmd_str);
	if (cmd == NULL) {
		printf("no such at command\n");
		return -1;
	}

	/* allocate memory to store the context for async cb */
	ctx = malloc(sizeof(*ctx));
	if (ctx == NULL) {
		printf("not enough memory\n");
		return -1;
	}

	memset(ctx, 0, sizeof(*ctx));
	ctx->type = ATCMD;
	ctx->atcmd_query = cmd;
	ctx->func = func;
	ctx->arg = arg;

	memcpy(&frame.atcmd.cmd, atcmd_str, 2);
	memcpy(&frame.buf, param, param_len);

	if (xbeeapp_send(ctx, XBEE_TYPE_ATCMD, &frame,
			 sizeof(struct xbee_atcmd_hdr) +
			 param_len, foreground) < 0) {
		free(ctx);
		return -1;
	}

	return 0;
}

/* dump config */
static int __xbeeapp_dump_config(struct xbee_ctx *ctx, struct xbee_atcmd *cmd)
{
	struct xbee_atcmd_hdr atcmd;

	/* find the first command that is readable and writable */
	for (; cmd->name != NULL; cmd++) {
		if ((cmd->flags & (XBEE_ATCMD_F_READ|XBEE_ATCMD_F_WRITE)) ==
		    (XBEE_ATCMD_F_READ|XBEE_ATCMD_F_WRITE))
			break;
	}
	if (cmd->name == NULL) {
		printf("no register to dump\n");
		return -1;
	}

	ctx->atcmd_query = cmd;
	ctx->have_more_command = 1;
	memcpy(&atcmd.cmd, cmd->name, 2);

	if (xbeeapp_send(ctx, XBEE_TYPE_ATCMD, &atcmd,
			 sizeof(struct xbee_atcmd_hdr), 1) < 0) {
		return -1;
	}

	return 0;
}

/* dump config */
int xbeeapp_dump_config(const char *filename)
{
	struct xbee_ctx *ctx;
	struct xbee_atcmd *cmd;
	int ret;

	/* find the first command that is readable and writable */
	cmd = &xbee_atcmd_list[0];

	/* allocate memory to store the context for async cb */
	ctx = malloc(sizeof(*ctx));
	if (ctx == NULL) {
		printf("not enough memory\n");
		return -1;
	}

	memset(ctx, 0, sizeof(*ctx));
	ctx->type = DUMP_CONF;
	ret = __xbeeapp_dump_config(ctx, cmd);

	if (ret < 0)
		free(ctx);
	return ret;
}

/* continue a dump conf */
static int xbeeapp_dump_conf_continue(struct xbee_ctx *ctx,
				      struct xbee_atresp_hdr *frame,
				      unsigned len)
{
	struct xbee_atcmd *cmd;
	char buf[32];

	/* see if it matches query */
	if (memcmp(&frame->cmd, ctx->atcmd_query->name, 2)) {
		printf("invalid response\n");
		ctx->have_more_command = 0;
		return -1;
	}

	/* dump register content in buf, skip on error */
	if (atcmd_frame_status(frame, len) == 0 &&
	    atcmd_frame_to_string(frame, len, buf, sizeof(buf)) == 0)
		printf("write %s %s\n", ctx->atcmd_query->desc, buf);

	cmd = ctx->atcmd_query + 1;
	if (__xbeeapp_dump_config(ctx, cmd) < 0) {
		ctx->have_more_command = 0;
		//close(xbee_config_file);
		printf("END\n"); /* XXX */
	}
	return 0;
}

/* send data message */
int xbeeapp_send_msg(uint64_t addr, void *data, unsigned data_len,
		     int foreground)
{
	struct xbee_ctx *ctx;
	struct {
		struct xbee_xmit_hdr xmit;
		char buf[XBEE_MAX_FRAME_LEN];
	} __attribute__((packed)) frame;

	/* allocate memory to store the context for async cb */
	ctx = malloc(sizeof(*ctx));
	if (ctx == NULL) {
		printf("not enough memory\n");
		return -1;
	}

	memset(ctx, 0, sizeof(*ctx));
	ctx->type = SEND_MSG;
	ctx->atcmd_query = NULL;

	frame.xmit.dstaddr = htobe64(addr);
	frame.xmit.reserved = htons(0xFFFE);
	frame.xmit.bcast_radius = 0;
	frame.xmit.opts = 0;
	memcpy(&frame.buf, data, data_len);

	if (xbeeapp_send(ctx, XBEE_TYPE_XMIT, &frame,
			 sizeof(struct xbee_xmit_hdr) +
			 data_len, foreground) < 0) {
		free(ctx);
		return -1;
	}

	return 0;
}

void xbee_stdin_enable(void)
{
	event_add(&stdin_read_event, NULL);
}

void xbee_stdin_disable(void)
{
	event_del(&stdin_read_event);
}

static void evt_timeout(int s, short event, void *arg)
{
	struct xbee_ctx *ctx = arg;

	xbeeapp_log(0, "Timeout\n");

	/* restart command line */
	if (ctx->foreground) {
		if (xbee_debug == 0)
			printf("Timeout\n");
		xbee_stdin_enable();
		rdline_newline(&xbee_cl->rdl, xbee_cl->prompt);
	}
	/* free event */
	xbee_unregister_channel(xbee_dev, ctx->channel);
	free(ctx);
}

void xbee_load_timeout(struct xbee_ctx *ctx)
{
	struct timeval tv;

	tv.tv_sec = 0;
	tv.tv_usec = TIMEOUT_US;
	evtimer_set(&ctx->timeout, evt_timeout, ctx);
	evtimer_add(&ctx->timeout, &tv);
}

void xbee_unload_timeout(struct xbee_ctx *ctx)
{
	evtimer_del(&ctx->timeout);
}

static void
evt_stdin_input(int s, short event, void *arg)
{
	char buf[2048];
	int n;

	n = read(s, buf, sizeof(buf));
	if (n < 0) {
		event_loopbreak();
		return;
	}

	if (xbee_raw) {
		int i;
		for (i = 0; i < n; i++) {
			/* ctrl-d is pressed, back in cmdline mode XXX */
			if (buf[i] == 4) {
				xbee_raw = 0;
				n = i;
				rdline_newline(&xbee_cl->rdl, xbee_cl->prompt);
				break;
			}
		}
		/* XXX bad hack */
		if (buf[0] == '\n')
			write(xbee_dev->fd, "\r\n", 2);
		else
			write(xbee_dev->fd, buf, n);
		write(s, buf, n);

	}
	else {
		if (cmdline_in(xbee_cl, buf, n) < 0) {
			event_loopbreak();
			return;
		}
	}
}

static void
evt_xbee_input(int s, short event, void *arg)
{
	if (xbee_raw) {
		char buf[2048];
		int n;

		n = read(s, buf, sizeof(buf));
		if (n < 0)
			return;

		write(1, buf, n);
	}
	else {
		xbee_read(xbee_dev);
		xbee_process_queue(xbee_dev);
	}
}

static void
usage(const char *prgname)
{
	printf("%s [-s BAUD] -d DEVNAME\n"
	       "  -s BAUD:      set baud rate of xbee device\n"
	       "  -d DEVNAME:   xbee char device\n"
	       "  -r:           start in raw mode\n"
	       "\n",
	       prgname);
}

/* Parse the argument given in the command line of the application */
static int
parse_args(int argc, char **argv)
{
	int opt, ret;
	char **argvopt;
	int option_index;
	char *prgname = argv[0];
	static struct option lgopts[] = {
		{0, 0, 0, 0}
	};

	argvopt = argv;

	while ((opt = getopt_long(argc, argvopt, "hd:s:j:",
				  lgopts, &option_index)) != EOF) {

		switch (opt) {

		case 'h':
			usage(prgname);
			exit(0);
			break;

		case 'd':
			xbee_devname = strdup(optarg);
			break;

		case 's':
			xbee_baud = atoi(optarg);
			break;

		case 'r':
			xbee_raw = 1;
			break;

		case 'j':
			joystick_devname = optarg;
			break;

		/* long options */
		case 0:
			/* if (!strcmp(lgopts[option_index].name, "option")) */
			break;

		default:
			usage(prgname);
			return -1;
		}
	}

	if (xbee_devname == NULL) {
		fprintf(stderr, "xbee device argument missing\n");
		usage(prgname);
		return -1;
	}

	if (argc != optind) {
		printf("Invalid argument\n");
		usage(prgname);
		return -1;
	}

	ret = optind-1;

	return ret;
}

int main(int argc, char **argv)
{
	struct termios oldterm, term;
	int err = 0;
	const char *homedir;
	char xbeerc_path[256];

	if (parse_args(argc, argv) < 0)
		exit(1);

	/* initializa libevent */
	event_init();

	/* initialize libxbee */
	err = xbee_init();
	if (err < 0)
		return -1;

	/* init rc_proto */
	rc_proto_init();

	/* init joystick */
	if (joystick_devname != NULL) {
		if (joystick_init(joystick_devname, &joyinfo) < 0) {
			fprintf(stderr, "error in joystick init\n");
			return -1;
		}
		event_set(&joystick_read_event, joyinfo.fd, EV_READ | EV_PERSIST,
			  joystick_input, &joyinfo);
		event_add(&joystick_read_event, NULL);
	}

	/* open xbee device */
	xbee_dev = xbee_open(xbee_devname, xbee_baud);
	if (xbee_dev == NULL)
		return -1;

	/* register default channel with a callback */
	if (xbee_register_channel(xbee_dev, XBEE_DEFAULT_CHANNEL,
				  xbee_rx, NULL) < 0) {
		fprintf(stderr, "cannot register default channel\n");
		return -1;
	}

	/* add read event on xbee device */
	event_set(&xbee_read_event, xbee_dev->fd, EV_READ | EV_PERSIST,
		  evt_xbee_input, xbee_dev);
	event_add(&xbee_read_event, NULL);

	/* set terminal in raw mode */
	tcgetattr(0, &oldterm);
	memcpy(&term, &oldterm, sizeof(term));
	term.c_lflag &= ~(ICANON | ECHO | ISIG);
	tcsetattr(0, TCSANOW, &term);
	setbuf(stdin, NULL);

	/* parse the config file */
	homedir = getenv("HOME");
	if (homedir != NULL) {
		snprintf(xbeerc_path, sizeof(xbeerc_path), "%s/.xbeerc",
			 homedir);
		if (access(xbeerc_path, R_OK) == 0) {

			xbee_cl = cmdline_file_new(&main_ctx, "xbeerc> ",
					      "/home/zer0/.xbeerc", 1);
			if (xbee_cl != NULL) {
				cmdline_interact(xbee_cl);
				cmdline_free(xbee_cl);
				printf("\nconfig file parsed\n");
			}
		}
	}

	/* create a new cmdline instance */
	xbee_cl = cmdline_stdin_new(&main_ctx, "xbee> ");
	if (xbee_cl == NULL) {
		err = 1;
		goto fail;
	}

	/* load read event on stdin */
	event_set(&stdin_read_event, 0, EV_READ | EV_PERSIST,
		  evt_stdin_input, xbee_cl);
	event_add(&stdin_read_event, NULL);

	/* libevent main loop */
	event_dispatch();

	tcsetattr(0, TCSANOW, &oldterm);
	return 0;

 fail:
	tcsetattr(0, TCSANOW, &oldterm);
	return err;
}
