/*
 * Copyright (c) 2011, Olivier MATZ <zer0@droids-corp.org>
 * All rights reserved.
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the University of California, Berkeley nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE REGENTS AND CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/queue.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <termios.h>

#include "xbee_neighbor.h"
#include "xbee_stats.h"
#include "xbee_buf.h"
#include "xbee_proto.h"
#include "xbee.h"

int xbee_init(void)
{
	return 0;
}

static int xbee_settermios(struct xbee_dev *dev)
{
	struct termios term;

	memset(&term, 0, sizeof(term));
	term.c_cflag = CREAD | HUPCL;

	switch (dev->baud) {
		case 50: term.c_cflag |= B50; break;
		case 75: term.c_cflag |= B75; break;
		case 110: term.c_cflag |= B110; break;
		case 134: term.c_cflag |= B134; break;
		case 150: term.c_cflag |= B150; break;
		case 200: term.c_cflag |= B200; break;
		case 300: term.c_cflag |= B300; break;
		case 600: term.c_cflag |= B600; break;
		case 1200: term.c_cflag |= B1200; break;
		case 1800: term.c_cflag |= B1800; break;
		case 2400: term.c_cflag |= B2400; break;
		case 4800: term.c_cflag |= B4800; break;
		case 9600: term.c_cflag |= B9600; break;
		case 19200: term.c_cflag |= B19200; break;
		case 38400: term.c_cflag |= B38400; break;
		case 57600: term.c_cflag |= B57600; break;
		case 115200: term.c_cflag |= B115200; break;
		case 230400: term.c_cflag |= B230400; break;
		default:
			fprintf(stderr, "invalid baudrate\n");
			return -1;
	}

	/* ignore modem control lines */
	term.c_cflag |= CLOCAL;

	/* non-canonical mode */
	term.c_lflag &= ~ICANON;

	/* minimum number of chars and minimum timeout for non-canonical read */
	term.c_cc[VMIN] = 1;
	term.c_cc[VTIME] = 0;

#if 0
	/* XXX later */
	switch (dev->parity) {
		case PARITY_ODD: term.c_cflag |= PARENB | PARODD; break;
		case PARITY_EVEN: term.c_cflag |= PARENB; break;
		default:		break;
	}

	if (twostopb)
		term.c_cflag |= CSTOPB;

	if (software)
		term.c_iflag |= IXON | IXOFF;
	if (hardware)
		term.c_cflag |= CRTSCTS;

#endif

	if (tcsetattr(dev->fd, TCSAFLUSH, &term) < 0) {
		fprintf(stderr, "ERROR: tcsetattr(TCSAFLUSH) failed: %s\n",
			strerror(errno));
		return -1;
	}
	return(0);
}

int xbee_register_channel(struct xbee_dev *dev, int channel,
			  xbee_rx_cb_t *rx_cb, void *opaque)
{
	/* user asked for any channel */
	if (channel == XBEE_CHANNEL_ANY) {
		int ch;

		/* skip XBEE_DEFAULT_CHANNEL == 0 */
		for (ch = 1; ch < XBEE_MAX_CHANNEL; ch++) {
			if (dev->channel[ch].registered == 0) {
				channel = ch;
				break;
			}
		}
		/* no available channels */
		if (channel == XBEE_CHANNEL_ANY)
			return -1;
	}
	/* user requested a specific channel */
	else if (channel < 0 || channel >= XBEE_MAX_CHANNEL ||
		 dev->channel[channel].registered == 1)
		return -1; /* not available */

	dev->channel[channel].registered = 1;
	dev->channel[channel].rx_cb = rx_cb;
	dev->channel[channel].arg = opaque;
	return channel;
}

int xbee_unregister_channel(struct xbee_dev *dev, int channel)
{
	if (channel < 0 || channel >= XBEE_MAX_CHANNEL ||
	    dev->channel[channel].registered == 0)
		return -1;
	dev->channel[channel].registered = 0;
	dev->channel[channel].rx_cb = NULL;
	dev->channel[channel].arg = NULL;
	return 0;
}

struct xbee_dev *xbee_open(const char *devname, unsigned baudrate)
{
	struct xbee_dev *dev;
	int fd;

	/* allocate structure for xbee device */
	dev = malloc(sizeof(*dev));
	if (dev == NULL) {
		fprintf(stderr, "not enough memory\n");
		return NULL;
	}
	memset(dev, 0, sizeof(*dev));

	/* open the serial port */
	fd = open(devname, O_NONBLOCK|O_RDWR);
	if (fd < 0) {
		fprintf(stderr, "cannot open %s: %s\n", devname,
			strerror(errno));
		free(dev);
		return NULL;
	}

	/* set termios */
	dev->baud = baudrate;
	dev->fd = fd;
	if (xbee_settermios(dev) < 0) {
		close(fd);
		free(dev);
		return NULL;
	}

	dev->name = strdup(devname);
	xbee_bufq_init(&dev->queue);
	xbee_neigh_init(dev);

	return dev;
}

/* read data from device fd and put it in queue */
int xbee_read(struct xbee_dev *dev)
{
	struct xbee_buf *xbuf;
	int n, tailroom, total = 0;
	char *data;

	do {
		/* get an xbuf to store data */
		xbuf = xbee_bufq_last(&dev->queue);
		if (xbuf == NULL || xbee_buf_tailroom(xbuf) == 0) {
			xbuf = xbee_buf_alloc();

			if (xbuf == NULL) {
				fprintf(stderr, "FATAL: cannot allocate buffer\n");
				return -1;
			}

			/* put the xbuf in queue */
			xbee_buf_enqueue(&dev->queue, xbuf);
		}

		/* read data from char device */
		tailroom = xbee_buf_tailroom(xbuf);
		data = xbee_buf_tail(xbuf);
		n = read(dev->fd, data, tailroom);

		/* error in read */
		if (n < 0) {
			fprintf(stderr, "read() failed: %s\n", strerror(errno));
			return -1;
		}

		/* update queue len and xbuf len */
		xbee_bufq_append(&dev->queue, n);

		total += n;

	} while (n == tailroom);

	return total;
}

/* process all data in queue */
int xbee_process_queue(struct xbee_dev *dev)
{
	char buf[XBEE_MAX_FRAME_LEN];
	int len, ret;

	while (1) {
		len = xbee_proto_get_frame(dev, buf, sizeof(buf));
		if (len == 0) /* no more frame */
			break;

		if (len < 0) {
			/* a frame was dropped... */
			fprintf(stderr, "xbee_proto_get_frame() failed\n");
			continue;
		}

		ret = xbee_proto_parse_frame(dev, buf, len);
		if (ret < 0) {
			fprintf(stderr, "xbee_proto_parse_frame() failed\n");
			continue;
		}
	}

	return 0;
}
